//
// Test code for Kemper LED Lamps connected to an Arduino board.
//
// This program assumes there are four push buttons connected to inputs 8 thru 11. The 
// buttons are used to ramp the LEDs up and down. Each button ramps one of the colors 
// on the KLL string. The ramp direction is flipped by pressing 8 and 11 together. The 
// Arduino status LED on the Arduino board is used to indicate ramp direction. 
// 
// This code is handy to test KLL being used as accent lighting. You can ramp in 
// different colors in different amounts to find the perfect color and intensity. Your 
// mileage my very. 
//
// Written By: Jim Kemp / Powerhouse Electronics.
// 
// ==================

#include "NewSoftSerial.h"      // Library for software defined comm port.
#include "KLL.h"                // Library for Kemper LED Lamps

// Set Pin 2 as both Tx and Rx - this pushes NewSoftSerial to use half duplex.
NewSoftSerial mySerial( 2, 2 );

#define COLORS 5
byte colors[COLORS] = { 
  A , R, G, B, W                // Defines an array we can index later.
};             

int pcbLED = 13;                // Port pin used by Arduino's own status LED.

int red, blu, grn, wht;
#define INC_VAL  1
#define DEC_VAL  0

#define NUM_NODES       10      // Maximum RAM set aside to hold node addresses.
byte iNodes[ NUM_NODES ];       // Define space to hold 10 node address.
byte iNodeNum = 0;              // Tracks how many nodes we've discovered on the bus.


// I think the Arduino calls this function automatically before "loop" starts.
// ==============================================================================
void setup()  
{
  pinMode( pcbLED, OUTPUT);     // sets the digital pin as output

  Serial.begin( 9600 );         // This is the debug comm port on the Arduino.

  // set the data rate for the NewSoftSerial port
  mySerial.begin( 9600 );       // KLL communicate at 9600 baud.
  
  // Make sure RAM is cleared.
  for ( int i = 0 ; i < NUM_NODES ; i++ ) iNodes[ i ] = 0;
  
  Serial.println( "Setup done." );
}


// ==============================================================================
// Arduino Main Loop
// ==============================================================================
void loop()                     // Main Loop - This is what Arduino calls it.
{
  bool dir = INC_VAL;
  byte i;
  
  lampReset( ALL_LAMPS );       // Global reset command sent to all lamps.
  setLevel( ALL_LAMPS, A, 0 );  // Set all colors on all lamps to zero.

  Serial.flush();               // Flush both debug port and KLL comm port.
  mySerial.flush();

  Serial.println( "Startup..." ); // Send a nice Hello World, we're alive message.
  turboPingPrint();             // Ask all nodes to report in.

  // Set four pins as input. To be used as button inputs later.
  for( i=8 ; i <= 11 ; i++ ) {
    pinMode( i, INPUT );
    digitalWrite( i, HIGH);     // Turn on pullup resistor.
  }


  // Start main forever loop here. 
  while( 1 ) {
    // Test if buttons 8 & 11 are both being held down together.
    if( ! digitalRead( 8 ) && ! digitalRead( 11 ) ) {
      // Flip the up/down direction flag.
      if ( dir ) dir = 0;
      else dir = 1;
      // Set the Arduino status LED so we know what direction we're sliding.
      if( dir ) digitalWrite( pcbLED, HIGH );
      else      digitalWrite( pcbLED, LOW );
      // Wait for the buttons to both be released.
      while( digitalRead( 8 ) || digitalRead( 11 ) );
      // Add some debounce delay.
      delay( 500 );
    }

    // Test if buttons 10 & 11 are both being held down together.
    if( ! digitalRead( 10 ) && ! digitalRead( 11 ) ) {
      // Wait for the buttons to both be released.
      while( digitalRead( 10 ) || digitalRead( 11 ) );
      // Add a little debounce delay.
      delay( 100 );
      
      // Call a function to dance the LEDs.
      danceLEDs();
    }


    // Test if button on input #8 is on.
    if( ! digitalRead( 8 ) ) {
      // Slide the red LED color.
      adj( &red, dir );
      setLevel( ALL_LAMPS, R, red );
    }

    // Do the same thing for green.
    if( ! digitalRead( 9 ) ) {
      adj( &grn, dir );
      setLevel( ALL_LAMPS, G, grn );
    }

    // Blue
    if( ! digitalRead( 10 ) ) {
      adj( &blu, dir );
      setLevel( ALL_LAMPS, B, blu );
    }

    // White
    if( ! digitalRead( 11 ) ) {
      adj( &wht, dir );
      setLevel( ALL_LAMPS, W, wht );
    }

    delay( 50 );        // Min loop time is 50ms.

    // Kemp the Arduino LED updated to the correct state.
    if( dir ) digitalWrite( pcbLED, HIGH );
    else      digitalWrite( pcbLED, LOW );
  }
}


// ==============================================================================
void danceLEDs( ) {
    byte d, dd;     // Used for local delay amount.
    byte i, j, c;   // Index counters.

    Serial.println( "Start Dance" );
    lampReset( ALL_LAMPS );       // Global reset command sent to all lamps.
    delay( 50 );

    if ( iNodeNum == 0 ) {
        Serial.println( "No nodes to dance with." );
        return;  // Abort if there are no lamps on the bus.
    }

    // Set all lamps and all colors to fastest ramp rate.
    setRampRate( ALL_LAMPS, A, 0x80, 0x80 );

    // Rip through all the nodes and flash each color in sequence.
    d = 50;    // Inner delay between each flash (in milliseconds).
    for ( j = 0 ; j < 3 ; j++ ) {      
        // Flash each color on each node quickly.
        for ( i = 0 ; i < iNodeNum ; i++ ) {
            for ( c = 1 ; c < COLORS ; c++ ) {
                pulseLED( iNodes[ i ], colors[ c ], 0xFF );     
                delay( d );
            }
        }
        delay( 250 );
    }

    // Bang a bunch of random LEDs.
    for ( j = 0 ; j < 150 ; j++ ) {      
        i = random( 0, iNodeNum );
        c = random( 1, COLORS );
        setLevel( iNodes[ i ], colors[ c ], 0xFF );     // On
        delay( 200 );
        setLevel( iNodes[ i ], colors[ c ], 0x00 );     // Off
        delay( 10 );
    }
    

    delay( 2000 );
    
    // Set all lamps and all colors to slowest ramp rate.
    setRampRate( ALL_LAMPS, A, 0x01, 0x01 );

    // Go through all the nodes and slowly ramp each color in sequence.
    d = 150;    // Inner delay between each flash (in milliseconds).
    dd = 4000;  // Longer delay.
    for ( j = 0 ; j < 3 ; j++ ) {      
        // Sequence through each node, one at a time.
        for ( i = 0 ; i < iNodeNum ; i++ ) {
            for ( c = 1 ; c < COLORS ; c++ ) {
                setLevel( iNodes[ i ], colors[ c ], 0xFF );     // Start ramp on one color.
                delay( d );
            }
            // Wait for all the ramps to finish.
            //delay( 2000 );
        }
        
        delay( dd );
        
        // Sequence through each node, one at a time again.
        for ( i = 0 ; i < iNodeNum ; i++ ) {
            for ( c = 1 ; c < COLORS ; c++ ) {
                setLevel( iNodes[ i ], colors[ c ], 0x00 );     // Start decay.
                delay( d );
            }
            // Wait for all the decays to finish.
            //delay( 2000 );
        }

        delay( dd );
    }

    // Using one color at a time, pulse all the nodes at the same time using the global command.
    for ( j = 0 ; j < 3 ; j++ ) {      
        for ( c = 1 ; c < COLORS ; c++ ) {
            pulseLED( ALL_LAMPS, colors[ c ], 0xFF );     
            delay( 1000 );
        }
        delay( 1000 );
    }

        
    delay( 50 );
    Serial.println( "End Dance" );
}

// Each LED on each node can have it's own ramp / decay rate. The ramp and decay 
// rates may be different. After reset, the nodes reset themselves to a decent 
// ramp / decay rate of 0x15.
// ==============================================================================
void setRampRate( byte node, byte leds, byte rmpUp, byte rmpDwn ) {
  byte buff[3];

  buff[0] = leds | SET_LED_RAMP_DECAY;
  buff[1] = rmpUp;
  buff[2] = rmpDwn;
  sndPacket( node, buff, 2 );
  delay( 30 );  
}


// After an "Adj" button is pressed on the Arduino, this function steps to the 
// light either up or down. The step amount is non-linear - this helps speed things up.
// Adjust the "val" based on the "dir" flag. For low values, the adjustment amount
// is also low. As the val gets bigger, the adjustment amount also gets bigger. This
// makes for a non-linear adjuster. That's good because the light output from an 
// LED is also non-linear.
// ==============================================================================
void adj( int *val, bool dir ) {
  int addr=1;

  if( *val > 50 ) addr = 2;
  if( *val > 100 ) addr = 4;
  if( *val > 150 ) addr = 8;
  if( *val > 200 ) addr = 16;

  if( dir ) (*val) += addr;
  else --(*val) -= addr;

  // Once we have peaked, turn off the Arduino LED so the user 
  // knows he's maxed out. Do the opposite when sliding down.
  if( *val > 255 ) { 
    *val = 255; 
    digitalWrite( pcbLED, LOW ); 
  }
  if( *val < 0 ) { 
    *val = 0; 
    digitalWrite( pcbLED, HIGH ); 
  }
}


// Send a pulse command to the KLLs.
// ==============================================================================
void pulseLED( byte node, byte leds, byte level ) {
  byte buff[2];

  buff[0] = leds | PULSE_LED;
  buff[1] = level;
  sndPacket( node, buff, 2 );
  delay( 3 );  
}

// Send a reset command to a KLL node. The node address can be the global node 
// address too.
// ==============================================================================
void lampReset( byte node ) {
  byte buff[1];

  buff[0] = RESET;
  sndPacket( node, buff, 1 );
  delay( 125 );         // Dwell for 125ms while all the nodes reset.
}

// This is an automated debug print routine. Call this routine to print the node 
// addresses of all the KLLs on the string. 
// ==============================================================================
void turboPingPrint( ) {
  byte i, n, len;

  turboPing( ALL_LAMPS );

  // The number of bytes in the fifo indicates how many nodes reported back.
  len =  mySerial.available();

  if ( len > 0 ) {
    Serial.print( "TurboPing found: " );
    Serial.println( len, DEC );         // Print number of nodes found.
    for( i = 0 ; i < len ; i++ ) {      // For each node, print the address.
      n = mySerial.read();
      if ( i < NUM_NODES ) {
        iNodes[ i ] = n;                // Save first x number of nodes in an array.
        iNodeNum = i+1;                 // Keep track of the total number of nodes.
      }
      Serial.print( i+1, DEC );
      Serial.print( " -> " );
      Serial.println( n, DEC );
    }
  }
  else {
    // Else, nobody responded to our ping!
    Serial.println( "TurboPing: no lamps found." );
  }
}

// Send the ping request out on the bus. Each node will respond in kind. To keep 
// nodes from stepping on each other, a node waits a fixed amount of time before 
// responding. The response delay is node_address * 2 in milliseconds. As an example,
// node 10 would wait 20ms before responding while node 11 would wait 22ms. So, max 
// wait time equals the highest node number * 2 on the bus. So, a string of 20 lamps, 
// with nodes numbered 1 to 20, would require 400ms for all nodes to report in.
// ==============================================================================
void turboPing( byte node ) {
  byte buff[1];

  mySerial.flush();
  buff[0] = START_TURBO_PING;
  sndPacket( node, buff, 1 );
  // All nodes will report back within 500ms. The results are keep in the serial
  // fifo buffer.
  delay( 500 );
}

// Send the command to set the light level on a node. Note, the "leds" variable 
// is a bit field. So, multiple LEDs can be set using the same command.
// ==============================================================================
void setLevel( byte node, byte leds, byte level ) {
  byte buff[2];

  buff[0] = leds | SET_LED_LEVEL;
  buff[1] = level;
  sndPacket( node, buff, 2 );
  delay( 3 );  
}

// This routine simply puts a packet on the KLL bus. Note, make sure your 
// byte array is big enough! Otherwise, this routine will index off the end of
// your array. You have been warned!
// All packets start with a node address and end with a chacksum byte. The number 
// data bytes depends on the command being sent. This routine had no notion of packet 
// length - this routine simply sends bytes based on the variable "len".
// ==============================================================================
void sndPacket( byte node, byte *s, byte len ) {
  byte i, cs;

  cs = node;                            // Init checksum to node address.
  mySerial.print( node, BYTE );         // Send node address.

  // Send bytes out on the bus. The number of bytes is set by "len".
  for ( i=0 ; i < len ; i++ ) {
    mySerial.print( s[i], BYTE );       // Send one byte out on the bus.
    cs += s[i];                         // Keep the checksum updated.
  }
  mySerial.print( cs & 0xFF, BYTE );    // Last byte on the bus is the checksum.
}


